from requests.adapters import HTTPAdapter
from requests_hardened.ip_filter import filter_host


class IPFilterAdapter(HTTPAdapter):
    def __init__(
        self,
        is_https_proto: bool,
        allow_loopback: bool = False,
        tls_sni_support: bool = True,
    ):
        """
        :param is_https_proto: whether it is HTTPS or insecure HTTP.
        :param allow_loopback: whether to allow loopback IP addresses in IP filtering.
        :param tls_sni_support: whether to add support for TLS SNI (enabled by default,
            and shouldn't be changed unless you are certain).
        """
        super().__init__()
        self._is_https_proto = is_https_proto
        self._allow_loopback = allow_loopback
        self._tls_sni_support = tls_sni_support

    def build_connection_pool_key_attributes(self, request, verify, cert=None):
        host_params, pool_kwargs = super().build_connection_pool_key_attributes(
            request, verify, cert
        )

        # Copy headers before mutating them as they may be a global variable used by
        # subsequent requests.
        # e.g., https://github.com/lepture/authlib/blob/a7d68b4c3b8a3a7fe0b62943b5228669f2f3dfec/authlib/oauth2/client.py#L205-L206
        if request.headers:
            request.headers = dict(**request.headers)
        else:
            request.headers = {}

        # Adds the original URL hostname as the 'Host' header.
        original_host = request.headers["Host"] = host_params["host"]

        # Set the original hostname for certificate validation otherwise
        # urllib3 will try to match the pinned resolved IP address against the
        # certificate.
        #
        # Note: assert_hostname and server_hostname cannot be passed
        #       when using insecure HTTP (http://) as it's not a valid argument for
        #       `urllib3.connectionpool.HTTPConnectionPool`.
        if self._is_https_proto is True:
            # For non-TLS SNI servers.
            pool_kwargs["assert_hostname"] = original_host  # type: ignore[typeddict-unknown-key] # Valid parameter for urllib3.connectionpool.HTTPSConnectionPool

            # Support TLS servers with SNI callbacks.
            if self._tls_sni_support is True:
                pool_kwargs["server_hostname"] = original_host  # type: ignore[typeddict-unknown-key] # Valid parameter for urllib3.connectionpool.HTTPSConnectionPool

        # Override the connection hostname to the resolved IP address,
        # and reject if it's a private IP.
        host_params["host"] = filter_host(
            original_host,
            host_params["port"],
            allow_loopback=self._allow_loopback,
        )
        return host_params, pool_kwargs

    def get_connection(self, url, proxies=None):
        # Note: we do not support this method due to being deprecated since May 2024.
        #       Only `get_connection_with_tls_context` is supported by our package,
        #       due to the deprecated `get_connection` being largely different
        #       and is unlikely to still be used by other packages.
        # Additional references:
        #   - https://github.com/psf/requests/pull/6655
        #   - https://github.com/psf/requests/pull/6710
        #   - https://github.com/advisories/GHSA-9wx4-h78v-vm56
        raise NotImplementedError(
            "get_connection is not supported in requests-hardened>=v1.0.0b5\n"
            "Upgrade your 'requests' package to >=2.32.2 and your dependencies "
            "if they rely on `requests.adapters.HTTPAdapter.get_connection`."
        )



from typing import Union, Any

import requests
from requests import Request, PreparedRequest, Response

from requests_hardened.config import Config
from requests_hardened.ip_filter_adapter import IPFilterAdapter

T_TIMEOUT = Union[int, float]


class HTTPSession(requests.Session):
    def __init__(self, config: Config, **kwargs):
        super().__init__(**kwargs)

        if config.ip_filter_enable is True:
            self.mount(
                "http://",
                IPFilterAdapter(
                    is_https_proto=False,  # We use http:// (insecure)
                    allow_loopback=config.ip_filter_allow_loopback_ips,
                    tls_sni_support=config.ip_filter_tls_sni_support,
                ),
            )
            self.mount(
                "https://",
                IPFilterAdapter(
                    is_https_proto=True,  # We use https:// (SSL/TLS)
                    allow_loopback=config.ip_filter_allow_loopback_ips,
                    tls_sni_support=config.ip_filter_tls_sni_support,
                ),
            )

        self._config = config

    def send(self, request: PreparedRequest, **kwargs: Any) -> Response:
        allow_redirects = kwargs.setdefault(
            "allow_redirects", not self._config.never_redirect
        )
        if allow_redirects and self._config.never_redirect:
            kwargs["allow_redirects"] = False

        timeout = kwargs.setdefault("timeout", self._config.default_timeout)
        if not timeout:
            kwargs["timeout"] = self._config.default_timeout
        return super().send(request, **kwargs)

    def prepare_request(self, request: Request) -> PreparedRequest:
        if self._config.user_agent_override is not None:
            request.headers.update({"User-Agent": self._config.user_agent_override})
        return super().prepare_request(request)



import ipaddress
import logging
import socket
from typing import Tuple, Union

import requests.exceptions
from urllib3.util.connection import (  # type: ignore[attr-defined] # `allowed_gai_family` exists. # noqa: E501
    allowed_gai_family,
)

logger = logging.getLogger(__name__)


class InvalidIPAddress(requests.RequestException):
    pass


def get_ip_address(
    hostname: str, port: int, allow_loopback: bool
) -> Tuple[Union[ipaddress.IPv4Address, ipaddress.IPv6Address], int]:
    # Development safeguard, hostname shouldn't be surrounded by brackets
    # when calling this function.
    assert not hostname.startswith("[")

    addresses = socket.getaddrinfo(
        hostname, port, family=allowed_gai_family(), type=socket.SOCK_STREAM
    )

    af, socktype, proto, _canonname, socket_address = addresses[0]

    if af != socket.AF_INET and af != socket.AF_INET6:
        # This code shouldn't be reachable.
        raise ValueError(
            "Only AF_INET and AF_INET6 socket address families are supported"
        )
    port = socket_address[1]
    ip = ipaddress.ip_address(socket_address[0])

    # Python considers special IP ranges representing IPv4 addresses as being
    # private ranges, such as ::ffff:0:0/96, ::/128.
    # Because of that, we need to check the IPv4 address instead of the IPv6 one.
    # https://github.com/python/cpython/commit/ed391090cc8332406e6225d40877db6ff44a7104
    if ip.version == 6 and (ipv4 := ip.ipv4_mapped) is not None:
        ip = ipv4

    if allow_loopback and ip.is_loopback:
        return ip, port
    elif ip.is_private:
        logger.warning(
            "Forbidden IP address: %s for hostname %s",
            ip,
            hostname,
            # Extra arguments may clash with logger configuration
            # if it injects extra JSON fields, such as:
            # https://github.com/saleor/saleor/blob/5e7c57dad9e64b09477ebdcee53f0277359bc598/saleor/core/logging.py#L13
            # Because of that, we prefix the extra arguments in order to
            # reduce the chance of clashing with logging configs.
            extra={"dst_ip": ip, "dst_hostname": hostname},
        )
        raise InvalidIPAddress(ip)
    return ip, port


def filter_host(
    hostname: str, port: int, *, allow_loopback: bool
) -> str:
    if not hostname:
        raise requests.exceptions.InvalidURL("Invalid URL: missing hostname")

    try:
        ip_addr, port = get_ip_address(
            hostname, port, allow_loopback=allow_loopback
        )
    except socket.gaierror as exc:
        raise requests.ConnectionError("Failed to resolve domain") from exc
    except socket.timeout as exc:
        raise requests.ConnectTimeout("Failed to connect to host") from exc

    ip_addr_str = str(ip_addr) if ip_addr.version != 6 else f"[{ip_addr}]"
    return ip_addr_str
